#!/usr/bin/python
# -*- coding: utf-8 -*-

from abc import ABC
import json
import os
import pandas as pd
import logging
import sys


# setup logger
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from .operator import IOperator


class COCO(IOperator, ABC):

    """ Instance Object for COCO annotation format """

    def __init__(self, dataset):
        super(COCO, self).__init__(dataset)
        self._dataset = dataset

    def extract(self, path: str):
        """
        all the annotations in the file convert into general dataframe object.
        :param path: string, relative / absolute path
        :return: generalize pandas.DataFrame type object.
        """
        if os.path.exists(path):
            with open(path) as fp:
                ann_data = json.load(fp)
            self.__updateDataset(ann_data["images"])
            self.__extractAnnotation(ann_data["annotations"])
            self.__extractClasses(ann_data["categories"])
        else:
            logger.error(f"\n ERROR : entered path <{path}> is invalid.")
            sys.exit(1)

        return

    def archive(self, location, data):
        """ save coco annotation file in the given location

        :param location: .json file saving directory
        :param data: dictionary to save as .json
        :return: none
        """
        if os.path.exists(os.path.dirname(location)):
            if data:
                with open(location, 'w') as pf:
                    json.dump(data, pf)
            else:
                logger.error("\n ERROR : The DataFrame file is empty.")
                sys.exit(1)
        else:
            logger.error("\n ERROR : There are no such parent directory to file save.")
            sys.exit(1)

    def translate(self):
        """ translate common schema into json compatible format.

        :return:
        """
        data = {}
        data["annotations"] = []
        data["images"] = []
        data["categories"] = []

        filenames = self._dataset["name"].to_list()
        widths = self._dataset["width"].to_list()
        heights = self._dataset["height"].to_list()
        ids = self._dataset["image_id"].to_list()
        compact_image_list = zip(filenames, heights, widths, ids)

        for line in compact_image_list:
            data["images"].append(self.__list2dict(['file_name', 'height', 'width', 'id'], line))

        obj_ids = self.annotations["obj_id"].astype('int32')
        image_ids = self.annotations["image_id"].astype('int32')
        cat_ids = self.annotations["class_id"].astype('int32')
        xmins = self.annotations["x_min"].astype('int32')
        ymins = self.annotations["y_min"].astype('int32')
        xmaxs = self.annotations["x_max"].astype('int32')
        ymaxs = self.annotations["y_max"].astype('int32')
        bboxs = []
        areas = []
        for i in range(len(xmaxs)):
            bboxs.append([xmins[i], ymins[i], xmaxs[i], ymaxs[i]])
            areas.append((xmaxs[i] - xmins[i]) * (ymaxs[i] - ymins[i]))

        compact_ann_list = zip(obj_ids, image_ids, cat_ids, areas, bboxs)
        for line in compact_ann_list:
            data["annotations"].append(
                self.__list2dict(["id", "image_id", "category_id", "area", "bbox", "ignore", "iscrowd"], line))

        class_ids = list(self.classes.keys())
        class_names = list(self.classes.values())
        compact_class_list = zip(class_ids, class_names)
        for line in compact_class_list:
            data["categories"].append(self.__list2dict(["id", "name", "supercategory"], line, padd='none'))

        return data

    def __normalized2KITTI(self, box):
        """

        :param box: [X, Y, width, height]
        :return: [(xmin, ymin), (xmax, ymax)]
        """
        o_x, o_y, o_width, o_height = box
        xmin = int(o_x - o_width / 2)
        ymin = int(o_y - o_height / 2)
        xmax = int(o_x + o_width / 2)
        ymax = int(o_y + o_height / 2)
        return [(xmin, ymin), (xmax, ymax)]

    def __updateDataset(self, images):
        """

        :param images: image attributes in the .json file
        :return: add id, image width & height columns to self.dataset
        """
        dataset_imgs = list(self._dataset.iloc[:, 0].values)
        ann_imgs = []
        ann_id = {}
        img_width = {}
        img_height = {}
        for obj in images:
            if obj["file_name"] in dataset_imgs:
                try:
                    ann_imgs.append(obj["file_name"])
                    ann_id[obj["file_name"]] = obj["id"]
                    img_width[obj["file_name"]] = obj["width"]
                    img_height[obj["file_name"]] = obj["height"]
                except Exception as error:
                    logger.exception("\n ERROR : annotation file doesn't in accept the format.")
                    sys.exit(1)
        if len(dataset_imgs) > len(ann_imgs):
            self._dataset = self._dataset.loc[self._dataset.loc[:, "name"].isin(ann_imgs), :]
            logger.warning("\nWARNING: all the images had not annotated!")
        self._dataset = self._dataset.copy()
        self._dataset["image_id"] = self._dataset["name"].map(ann_id)
        self._dataset.loc[:, "width"] = self._dataset.loc[:, "name"].map(img_width)
        self._dataset.loc[:, "height"] = self._dataset.loc[:, "name"].map(img_height)
        super(COCO, self).set_dataset(self._dataset)
        return

    def __extractAnnotation(self, anns):
        """

        :param anns: annotation attribute in the .json file
        :return: None , add self.annotations attr.
        """
        ann_list = []
        for obj in anns:
            try:
                obj_id = obj["id"]
                img_id = obj["image_id"]
                cls_id = obj["category_id"]
                min_tup, max_tup = self.__normalized2KITTI(obj["bbox"])
            except Exception as error:
                logger.exception("\n ERROR : annotation file doesn't in accept the format.")
                sys.exit(1)
            else:
                ann_list.append((obj_id, img_id, cls_id, min_tup[0], min_tup[1], max_tup[0], max_tup[1]))

        if ann_list:
            ann_df = pd.DataFrame.from_records(ann_list,
                                               columns=['obj_id', 'image_id', 'class_id', 'x_min', 'y_min', 'x_max',
                                                        'y_max'])
            super(COCO, self).set_annotations(ann_df)
        else:
            super(COCO, self).set_annotations(pd.DataFrame())

    def __extractClasses(self, cats):
        """

        :param cats: categories attribute in the .json file
        :return: dictionary object of type {id : class-name}
        """
        if len(cats) > 0:
            class_dict = {}
            for obj in cats:
                try:
                    class_dict[obj["id"]] = obj["name"]
                except Exception as error:
                    logger.exception("\n ERROR : annotation file doesn't in accept the format.")
                    sys.exit(1)

            super(COCO, self).set_classes(class_dict)
        else:
            logger.error("\n ERROR : There are no distinctive class definition in the annotation.")
            sys.exit(1)
            super(COCO, self).set_classes({})

    def __KITTI2normilized(self, xmin, ymin, xmax, ymax):
        """

        :param xmin:
        :param ymin:
        :param xmax:
        :param ymax:
        :return: [X, Y, width, height]
        """
        width = xmax - xmin
        height = ymax - ymin
        x0 = (xmax + xmin) / 2
        y0 = (ymin + ymax) / 2
        return [x0, y0, width, height]

    def __list2dict(self, tags, values, padd='0'):
        nt = len(tags)
        nv = len(values)
        if nt >= nv:
            ret_dict = {}
            for i in range(nt):
                if i < nv:
                    #ret_dict[tags[i]] = str(values[i])
                    if isinstance(values[i], int):
                        ret_dict[tags[i]] = int(values[i])
                    elif isinstance(values[i], list):
                        cl_list = [int(elem) for elem in values[i]]
                        ret_dict[tags[i]] = cl_list
                    else:
                        ret_dict[tags[i]] = str(values[i])
                else:
                    ret_dict[tags[i]] = padd
            else:
                return ret_dict
        else:
            logger.error(f"\n ERROR : There are not enough attributes to create .json file.\n #tags : {nt} & #attrs : {nv}")
            sys.exit(1)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from abc import ABC
import logging
import os
import sys
import pandas as pd

# setup logger
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from .operator import IOperator


class CSV(IOperator, ABC):

    """ Instance Object for COCO annotation format """

    def __init__(self, dataset):
        super().__init__(dataset)
        self._dataset = dataset
        self.attrs = ['filename', 'width', 'height', 'class', 'xmin', 'ymin', 'xmax', 'ymax']

    def extract(self, path: str):
        """
        all the annotations in the file convert into general dataframe object.
        :param path: string, relative / absolute path
        :return: generalize pandas.DataFrame type object.
        """
        if os.path.exists(path):
            ann_df = pd.read_csv(path)
            attr_df_list = list(ann_df.columns)
            if all(x in attr_df_list for x in self.attrs):
                new_ann_df = self.__dfUpdates(ann_df)
                self.__updateDataset(new_ann_df.loc[:, ["name", "width", "height", "image_id"]])
                self.__setAnn(new_ann_df)
            else:
                logger.error(f"\n ERROR : Entered annotation file does not contains all the required attributes. \n {self.attrs}")
                sys.exit()
                
        else:
            logger.error(f"\n ERROR : Entered directory {path}, does not exsist.")
            sys.exit()

    def archive(self, location, df):
        """ save csv annotation file in the given location

        :param location: .csv file saving location
        :param df: finalized DataFrame object from the self.translate()
        :return: None
        """
        if os.path.exists(os.path.dirname(location)):
                df.to_csv(location, index=False)
        else:
            logger.error("\n ERROR : There are no such parent directory to file save.")
            sys.exit(1)

    def translate(self):
        """ translate common schema into csv compatible format.

        :return: pd.DataFrame object with ["filename", "width", "height", "class", "xmin", "ymin", "xmax", "ymax"] columns.
        """
        csv_ann_df = self.annotations.copy()

        class_series = pd.Series(self.classes)
        csv_ann_df.loc[:,"class"] = csv_ann_df["class_id"].map(class_series)

        filename_series = pd.Series(self._dataset['name'].tolist(),
                                    index=self._dataset["image_id"].tolist())
        csv_ann_df.loc[:,"filename"] = csv_ann_df["image_id"].map(filename_series)

        width_series = pd.Series(self._dataset['width'].tolist(),
                                    index=self._dataset["image_id"].tolist())
        csv_ann_df.loc[:, "width"] = csv_ann_df["image_id"].map(width_series)

        height_series = pd.Series(self._dataset['height'].tolist(),
                                    index=self._dataset["image_id"].tolist())
        csv_ann_df.loc[:, "height"] = csv_ann_df["image_id"].map(height_series)

        if (pd.isnull(csv_ann_df["class"]).sum() + pd.isnull(csv_ann_df["filename"]).sum()) != 0:
            logger.error(f"\n ERROR : There are not enough data in past annotation file to create annotation file. {pd.isnull(csv_ann_df['class']).sum()}, {pd.isnull(csv_ann_df['filename']).sum()}")
            sys.exit(1)
        else:
            csv_ann_df.rename(columns={"x_min": "xmin", "y_min": "ymin", "x_max": "xmax", "y_max": "ymax"},
                              inplace=True)
            return csv_ann_df.loc[:, ["filename", "width", "height", "class", "xmin", "ymin", "xmax", "ymax"]]

    def __dfUpdates(self, full_df):
        """add id, image width & height columns to self.dataset

        :param full_df: read .csv file from annotation file.
        :return: refine DataFrame object with column of ["name" ,"obj_id", "image_id", "class_id", "x_min", "y_min", "x_max", "y_max"]
        """
        full_df = full_df.copy()

        uni_files = list(full_df.loc[:, "filename"].unique())
        ns = len(uni_files)
        file_id_col = pd.Series(range(1, ns + 1), index=uni_files)
        full_df["image_id"] = full_df["filename"].map(file_id_col)

        cats = list(full_df.loc[:, "class"].unique())
        nc = len(cats)
        cat_id_col = pd.Series(range(1, nc + 1), cats)
        full_df["class_id"] = full_df["class"].map(cat_id_col)

        self.__defineClasses(nc, cats)
        full_df.drop("class", inplace=True, axis=1)

        full_df["obj_id"] = pd.Series(range(1, full_df.shape[0] + 1))
        full_df.rename(columns={"filename": "name", "xmin": "x_min", "ymin": "y_min", "xmax": "x_max", "ymax": "y_max"},
                       inplace=True)

        return full_df

    def __defineClasses(self, n_ids, classes):
        """

        :param n_ids: number of unique classes
        :param classes: class list
        :return: dictionary object of type {id : class-name}
        """
        if n_ids == len(classes):
            ids = range(1, n_ids + 1)
            super(CSV, self).set_classes(dict(zip(ids, classes)))
        else:
            logger.error(f"\nlength of class names[{len(classes)}] and class ids[{n_ids}] are not equal.")
            sys.exit(1)

    def __setAnn(self, full_df):
        """

        :param full_df: refined DataFrame object.
        :return: set generalized annotation df object as self.annotations
        """
        full_df = full_df.copy()
        col_lis = ["obj_id", "image_id", "class_id", "x_min", "y_min", "x_max", "y_max"]
        if all(y in list(full_df.columns) for y in col_lis):
            ann_df = full_df.loc[:, col_lis]
            super(CSV, self).set_annotations(ann_df)
        else:
            logger.error(f"\nthere are missing of required columns in {full_df.columns}")
            sys.exit(1)

    def __updateDataset(self, image_df):
        """

        :param image_df: image attributes DataFrame
        :return: merge current self.__dataset with image_df.
        """
        partial_df = image_df.copy()
        res_df = pd.merge(self._dataset, partial_df, on="name")
        super(CSV, self).set_dataset(res_df)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from abc import ABC
import xml.etree.ElementTree as ET
import pandas as pd
import os
import re
import sys
import logging

# setup logger
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from .operator import IOperator


class PascalVOC(IOperator, ABC):

    """ Instance Object for pascal VOC annotation format """

    def __init__(self, dataset):
        super().__init__(dataset)
        self._dataset = dataset

    def extract(self, path: str):
        """ extract annotation data when input the path to .xml files

        :param path: string, relative / absolute path for annotation folder
        :return:
        """
        files_list = self.__extractFiles(path)
        image_id = 0
        img_list = []
        tol_obj_list = []
        for file in files_list:
            img_data, obj_list = self.__FileReader(os.path.abspath(path) + os.sep + file)
            image_id += 1
            img_data.append(image_id)
            obj_list = [i+[image_id] for i in obj_list]
            img_list.append(img_data)
            tol_obj_list.extend(obj_list)
        if img_list:
            img_df = pd.DataFrame.from_records(img_list, columns=['name', 'width', 'height','image_id'])
            self.__updateDataset(img_df)
        else:
            logger.error("[var]: img_list is empty.")
            sys.exit(1)

        if obj_list and len(obj_list[0]) == 6:
            obj_df = pd.DataFrame.from_records(tol_obj_list,
                                               columns=['x_min', 'y_min', 'x_max', 'y_max', 'class', 'image_id'])
            self.__DFRefiner(obj_df)
        else:
            logger.error(f"\n ERROR : obj_list has not many attrs. : {len(obj_list[0])} or obj_list is empty : {len(obj_list)}")
            sys.exit(1)

    def archive(self, location: str, data):
        """ save pascalVOC annotation file in the given location

        :param location: .xml file saving location
        :param data: .xml daa bundle
        :return:
        """
        try:
            tree_str = ET.tostring(data)
            with open(location, 'wb') as pf:
                pf.write(tree_str)
        except Exception as error:
            logger.exception(error)
            sys.exit(1)

    def translate(self):
        """ translate common schema into json compatible format.

        :return: none
        """
        for index, row in self._dataset.iterrows():
            ann_list = self.__filterImgObj(row['image_id'])
            box = self.__xmlFormatter(row, ann_list)
            if box:
                yield box, row['name']
            else:
                yield None

    def __extractFiles(self, path: str):
        """

        :param path: relative or absolute directory to the annotation folder.
        :return: return list of all .xml file names in given directory.
        """
        if os.path.exists(path):
            if not [x[1] for x in os.walk(path) if x[1] != []]:
                path_list = [y[2] for y in os.walk(path) if y[2] != []][0]
                if path_list:
                    xml_list = [n for n in path_list if n.split('.')[-1] == 'xml']
                    if xml_list:
                        return xml_list
                    else:
                        logger.error("\n ERROR : There are no .xml files in the given directory.")
                        sys.exit(1)
                else:
                    logger.error("\n ERROR : The folder is empty.")
                    sys.exit(1)
        else:
            logger.error(f"\n ERROR : The entered path <{path}> is not valid.")
            sys.exit(1)

    def __DFRefiner(self, ann_df):
        """
        create pd.DataFrame with columns of [ "obj_id", "image_id", "class_id", "x_min", "y_min", "x_max", "y_max" ] and
        define self.annotations and self.classes
        :param ann_df: pd.Dataframe with columns of [ 'x_min', 'y_min', 'x_max', 'y_max', 'class', 'image_id' ]
        :return: None
        """
        ann_df = ann_df.copy()

        cats = list(ann_df.loc[:, "class"].unique())
        n_cats = len(cats)
        cat_series = pd.Series(range(1, n_cats + 1), index=cats)

        ann_df["class_id"] = ann_df["class"].map(cat_series)
        ann_df["obj_id"] = range(1,ann_df.shape[0]+1)
        nw_df = ann_df.loc[:, ["obj_id", "image_id", "class_id", "x_min", "y_min", "x_max", "y_max"]]

        super(PascalVOC, self).set_annotations(nw_df)
        super(PascalVOC, self).set_classes(dict(zip(range(1,n_cats+1),cats)))

    def __FileReader(self, file_path: str):
        """ read individual xml files extract data, create pd.DataFrame files

        :param file_path: absolute path to the single .xml file
        :return: tuple of two list
         img_data = [ filename, width, height ]
         obj_list = [ class, xmin, ymin, xmax, ymax ]
        """
        ann_tree = ET.parse(file_path)
        ann_root = ann_tree.getroot()
        try:
            filename = self.__tagFilter(ann_root.find('filename').text)
            size = ann_root.find('size')
            width = int(size.find('width').text)
            height = int(size.find('height').text)
            img_data = [filename, width, height]

            obj_list = []
            for obj in ann_root.findall('object'):
                obj_list.append(self.__get_voc_annotation_from_obj(obj))
        except Exception as error:
            logger.exception(error)
            sys.exit(1)

        return [img_data, obj_list]

    def __get_voc_annotation_from_obj(self, obj):
        """ read <object> block in xml file

        :param obj: <object> block in the .xml file
        :return: a list of object attrs. [ class, xmin, ymin, xmax, ymax ]
        """
        try:
            label = self.__tagFilter(obj.find('name').text)
            bndbox = obj.find('bndbox')
            xmin = int(bndbox.find('xmin').text)
            ymin = int(bndbox.find('ymin').text)
            xmax = int(bndbox.find('xmax').text)
            ymax = int(bndbox.find('ymax').text)
            ann = [xmin, ymin, xmax, ymax, label]
            return ann
        except Exception as error:
            logger.exception(error)
            sys.exit(1)

    def __updateDataset(self, image_df):
        """

        :param image_df: image attributes DataFrame
        :return: merge current self.__dataset with image_df.
        """
        partial_df = image_df.copy()
        res_df = pd.merge(self._dataset, partial_df, on="name")
        super(PascalVOC, self).set_dataset(res_df)

    def __filterImgObj(self, img_id):
        """ get row for specific image_id.

        :param img_id: [int] image_id in the self.annotations
        :return: all the row that carry given image_id as a list.
        """
        filtered_list = self.annotations.loc[self.annotations["image_id"] == int(img_id), :].values.tolist()
        return filtered_list

    def __xmlFormatter(self, image_data, ann_data):
        """ build the structure of the .xml file with data.

        :param image_data: dictionary for data in self._dataset
        :param ann_data: return values from self.__filterImgObj
        :return: complete .xml object
        """
        try:
            ann = ET.Element("annotation")
            ET.SubElement(ann, 'folder').text = image_data['folder']
            ET.SubElement(ann, 'filename').text = image_data['name']
            ET.SubElement(ann, 'path').text = image_data['path']
            size = ET.SubElement(ann, 'size')
            ET.SubElement(size, 'width').text = str(image_data['width'])
            ET.SubElement(size, 'height').text = str(image_data['height'])
            ET.SubElement(size, 'depth').text = str(3)
            for line in ann_data:
                obj = ET.SubElement(ann, 'object')
                ET.SubElement(obj, 'name').text = self.classes[line[2]]
                ET.SubElement(obj, 'pose').text = 'Unspecified'
                ET.SubElement(obj, 'truncated').text = str(0)
                ET.SubElement(obj, 'difficult').text = str(0)
                bbox = ET.SubElement(obj, 'bndbox')
                ET.SubElement(bbox, 'xmin').text = str(line[3])
                ET.SubElement(bbox, 'ymin').text = str(line[4])
                ET.SubElement(bbox, 'xmax').text = str(line[5])
                ET.SubElement(bbox, 'ymax').text = str(line[6])
            return ann
        except Exception as error:
            logger.exception(error)
            sys.exit(1)

    def __tagFilter(self,st: str):
        s = re.sub(r'\t*\n*\r*', '', st)
        return s




from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(

    name='imgann',

    version='0.7.5',

    description='image annotation file operation provider.',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/nipdep/imgann.git',

    author='nipdep',

    author_email='nipun1deelaka@gmail.com',

    # classifiers=[  # Optional
    #
    #     'Development Status :: 3 - Alpha',
    #
    #     'Intended Audience :: Developers',
    #     'Topic :: Software Development :: Build Tools',
    #
    #     'License :: OSI Approved :: MIT License',
    #
    #     'Programming Language :: Python :: 3',
    #     'Programming Language :: Python :: 3.5',
    #     'Programming Language :: Python :: 3.6',
    #     'Programming Language :: Python :: 3.7',
    #     'Programming Language :: Python :: 3.8',
    #     'Programming Language :: Python :: 3 :: Only',
    # ],


    keywords='sample, setuptools, development',

    # package_dir={'': 'src'},

    # py_modules = ["helloworld"],
    packages=['imgann','imgann/operators'],

    python_requires='>=3.6, <4',

    install_requires=['numpy', 'pandas', 'matplotlib', 'opencv-python'],

    extras_require={
        'dev': ['check-manifest',
                'pytest>=3.7'],
        'test': ['coverage', 'unittest'],
    },

    # package_data={  # Optional
    #     'sample': ['package_data.dat'],
    # },

    # data_files=[('my_data', ['data/data_file'])],  # Optional

    # entry_points={  # Optional
    #     'console_scripts': [
    #         'hello=sample:main',
    #     ],
    # },

    project_urls={  # Optional
        'Bug Reports': 'https://github.com/nipdep/imgann/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/nipdep/imgann/',
    },
)
